/*
 *  jDTAUS - DTAUS fileformat.
 *  Copyright (c) 2005 Christian Schulte <cs@schulte.it>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
package org.jdtaus.core.container.ri.client;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.jdtaus.core.container.Dependencies;
import org.jdtaus.core.container.Dependency;
import org.jdtaus.core.container.DuplicateImplementationException;
import org.jdtaus.core.container.Implementation;
import org.jdtaus.core.container.Implementations;
import org.jdtaus.core.container.InheritanceConstraintException;
import org.jdtaus.core.container.MissingImplementationException;
import org.jdtaus.core.container.Model;
import org.jdtaus.core.container.ModelError;
import org.jdtaus.core.container.Module;
import org.jdtaus.core.container.Modules;
import org.jdtaus.core.container.MultiplicityConstraintException;
import org.jdtaus.core.container.Properties;
import org.jdtaus.core.container.Property;
import org.jdtaus.core.container.PropertyException;
import org.jdtaus.core.container.PropertyOverwriteConstraintException;
import org.jdtaus.core.container.Specification;
import org.jdtaus.core.container.Specifications;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * {@code Model} reference implementation.
 *
 * @author <a href="mailto:cs@schulte.it">Christian Schulte</a>
 * @version $Id: DefaultModel.java 4342 2007-08-23 01:51:37Z schulte2005 $
 *
 * @see <a href="../../../../../../../jdtaus-module-1.1.xsd">jdtaus-module-1.1.xsd</a>
 */
public class DefaultModel implements Model
{
    //--Constants---------------------------------------------------------------

    /** JAXP configuration key to the Schema implementation attribute. */
    public static final String SCHEMA_LANGUAGE_KEY =
        "http://java.sun.com/xml/jaxp/properties/schemaLanguage";

    /** JAXP Schema implementation to use. */
    public static final String SCHEMA_LANGUAGE =
        "http://www.w3.org/2001/XMLSchema";

    /** JAXP configuration key for setting the Schema source. */
    public static final String SCHEMA_SOURCE_KEY =
        "http://java.sun.com/xml/jaxp/properties/schemaSource";

    /** jDTAUS {@code Model} namespace URI. */
    public static final String MODEL_NS =
        "http://jdtaus.org/runtime/container/xml";

    /** Location of the XML files to load. */
    public static final String MODEL_LOCATION =
        "META-INF/jdtaus/module.xml";

    /** Location of the jdtaus.xsd schema. */
    public static final String MODEL_XSD =
        "org/jdtaus/core/container/xml/jdtaus-module-1.1.xsd";

    /** Versions supported by this implementation. */
    private static final String SUPPORTED_VERSIONS[] = { "1.0", "1.1" };

    /** String constructor arguments. */
    private static final Class[] CTOR_ARGS_STRING = { String.class };

    /** String constructor arguments. */
    private static final Class[] CTOR_ARGS_CHAR = { char.class };

    //---------------------------------------------------------------Constants--
    //--Constructors------------------------------------------------------------

    /**
     * Creates a new {@code DefaultModel} instance.
     *
     * @throws ModelError if no model can be read.
     *
     * @see #readModules()
     */
    public DefaultModel()
    {
        try
        {
            this._modules = this.readModules();
        }
        catch (ParserConfigurationException e)
        {
            throw new ModelError(e);
        }
        catch (SAXException e)
        {
            throw new ModelError(e);
        }
        catch (IOException e)
        {
            throw new ModelError(e);
        }
    }

    //------------------------------------------------------------Constructors--
    //--Model-------------------------------------------------------------------

    /** Holds the loaded model. */
    private final Modules _modules;

    public final Modules getModules()
    {
        return this._modules;
    }

    //-------------------------------------------------------------------Model--
    //--DefaultModel------------------------------------------------------------

    /**
     * Gets the classloader used for loading classes.
     * <p>The reference implementation will use the current thread's context
     * classloader and will fall back to the system classloader if the
     * current thread has no context classloader set.</p>
     *
     * @return the classloader to be used for loading classes.
     *
     * @throws ModelError if no classloader is available.
     */
    protected ClassLoader getClassLoader()
    {
        ClassLoader classLoader = Thread.currentThread().
            getContextClassLoader();

        if(classLoader == null)
        {
            classLoader = ClassLoader.getSystemClassLoader();
        }

        if(classLoader == null)
        {
            throw new ModelError(new NullPointerException(
                "classLoader"));

        }

        return classLoader;
    }

    /**
     * Reads the module resources.
     * <p>The reference implementation reads all
     * {@code "META-INF/jdtaus/module.xml"} XML documents from the classpath
     * returning an instance of {@code Modules} holding theire data.</p>
     *
     * @throws IOException if reading fails.
     * @throws ParserConfigurationException if no working XML parser is
     * available supporting the needs of the implementation.
     * @throws SAXException if parsing the XML documents fails.
     */
    protected Modules readModules() throws IOException,
        ParserConfigurationException, SAXException
    {
        Document doc;
        final Collection documents = new LinkedList();
        final DocumentBuilder xmlBuilder;
        final DocumentBuilderFactory xmlFactory =
            DocumentBuilderFactory.newInstance();

        final Enumeration resources = this.getClassLoader().
            getResources(DefaultModel.MODEL_LOCATION);

        xmlFactory.setNamespaceAware(true);
        try
        {
            xmlFactory.setValidating(true);
            xmlFactory.setAttribute(DefaultModel.SCHEMA_LANGUAGE_KEY,
                DefaultModel.SCHEMA_LANGUAGE);

            final URL schema = this.getClassLoader().getResource(
                DefaultModel.MODEL_XSD);

            xmlFactory.setAttribute(DefaultModel.SCHEMA_SOURCE_KEY,
                schema.openStream());

        }
        catch(IllegalArgumentException e)
        {
            Logger.getLogger(DefaultModel.class.getName()).
                log(Level.CONFIG, e.getMessage());

            Logger.getLogger(DefaultModel.class.getName()).log(Level.WARNING,
                DefaultModelBundle.getNoValidationWarningMessage(
                Locale.getDefault()).format(new Object[] { e.getMessage() }));

            xmlFactory.setValidating(false);
        }

        xmlBuilder = xmlFactory.newDocumentBuilder();
        xmlBuilder.setErrorHandler(new ErrorHandler()
        {
            public void warning(final SAXParseException e)
            {
                Logger.getLogger(DefaultModel.class.getName()).
                    log(Level.WARNING, e.getMessage());

            }
            public void fatalError(final SAXParseException e)
            {
                throw new ModelError(e);
            }
            public void error(final SAXParseException e)
            {
                throw new ModelError(e);
            }
        });

        // Parse and validate all files from the classpath.
        while(resources.hasMoreElements())
        {
            final URL url;
            InputStream stream = null;

            try
            {
                url = (URL) resources.nextElement();
                stream = url.openStream();
                doc = xmlBuilder.parse(stream);
                documents.add(doc);
            }
            finally
            {
                if(stream != null)
                {
                    stream.close();
                    stream = null;
                }
            }
        }

        return this.transformModules(documents);
    }

    /**
     * Resolves instances for identifiers.
     *
     * @param modules the modules for which to resolve instances for identfiers.
     *
     * @throws NullPointerException if {@code modules} is {@code null}.
     */
    protected void linkXMLModel(final Modules modules)
    {
        if(modules == null)
        {
            throw new NullPointerException("modules");
        }

        Collection col;
        Dependencies deps;
        Implementations impls;
        Iterator it;
        Map map;
        Specification spec;
        Specifications specs;
        XMLDependency dep;
        XMLImplementation impl;
        XMLImplementation parent;
        final Map specsToImpls = new HashMap();
        final List unresolved = new LinkedList();

        // Link implementations.
        impls = modules.getImplementations();
        for(int i = impls.size() - 1; i >= 0; i--)
        {
            impl = (XMLImplementation) impls.getImplementation(i);
            unresolved.add(impl.getIdentifier());
        }

        while(unresolved.size() > 0)
        {
            impl = parent = (XMLImplementation) impls.getImplementation(
                (String) unresolved.get(unresolved.size() - 1));

            // Link the highest unresolved parent first.
            do
            {
                if(parent.getParentIdentifier() != null &&
                    parent.getParentIdentifier().length() > 0)
                {
                    parent = (XMLImplementation) impls.getImplementation(
                        parent.getParentIdentifier());

                    if(unresolved.contains(parent.getIdentifier()))
                    {
                        impl = parent;
                    }
                }
            } while(parent.getParentIdentifier() != null &&
                parent.getParentIdentifier().length() > 0);

            unresolved.remove(impl.getIdentifier());

            // Build specification mapping.
            String[] implemented = impl.getImplementedIdentifiers();
            specs = new Specifications();
            col = new ArrayList(implemented.length);
            for(int j = implemented.length - 1; j >= 0; j--)
            {
                if((map = (Map) specsToImpls.get(implemented[j])) == null)
                {
                    map = new HashMap();
                    specsToImpls.put(implemented[j], map);
                }

                if(map.put(impl.getName(), impl) != null)
                {
                    throw new DuplicateImplementationException(
                        impl.getName());

                }

                col.add(modules.getSpecifications().
                    getSpecification(implemented[j]));

            }

            specs.setSpecifications((Specification[]) col.
                toArray(new Specification[col.size()]));

            impl.setImplementedSpecifications(specs);

            if(impl.getParentIdentifier() != null &&
                impl.getParentIdentifier().length() > 0)
            {
                parent = (XMLImplementation) impls.
                    getImplementation(impl.getParentIdentifier());

                if(parent == null)
                {
                    throw new MissingImplementationException(
                        impl.getParentIdentifier());

                }
                else if(parent.isFinal())
                {
                    throw new InheritanceConstraintException(
                        impl.getIdentifier());

                }

                impl.setParent(parent);
            }
            else
            {
                impl.setParent(null);
            }

            // Update specification mapping for inherited specifications.
            specs = impl.getImplementedSpecifications();
            for(int i = specs.size() - 1; i >= 0; i--)
            {
                boolean alreadyMapped = false;
                spec = specs.getSpecification(i);
                implemented = impl.getImplementedIdentifiers();

                for(int j = implemented.length - 1; j >= 0; j--)
                {
                    if(implemented[j].equals(spec.getIdentifier()))
                    {
                        alreadyMapped = true;
                        break;
                    }
                }

                if(!alreadyMapped)
                {
                    if((map = (Map) specsToImpls.get(
                        spec.getIdentifier())) == null)
                    {
                        map = new HashMap();
                        specsToImpls.put(spec.getIdentifier(), map);
                    }

                    if(map.put(impl.getName(), impl) != null)
                    {
                        throw new DuplicateImplementationException(
                            impl.getName());

                    }
                }
            }
        }

        // Link specifications.
        specs = modules.getSpecifications();
        for(int i = specs.size() - 1; i >= 0; i--)
        {
            col = new LinkedList();
            spec = specs.getSpecification(i);
            impls = new Implementations();
            map = (Map) specsToImpls.get(spec.getIdentifier());
            if(map != null)
            {
                for(it = map.keySet().iterator(); it.hasNext();)
                {
                    col.add((XMLImplementation) map.get(it.next()));
                }

                impls.setImplementations((Implementation[]) col.
                    toArray(new Implementation[col.size()]));

            }
            spec.setImplementations(impls);
        }

        // Link dependencies.
        impls = modules.getImplementations();
        unresolved.clear();
        for(int i = impls.size() - 1; i >= 0; i--)
        {
            impl = (XMLImplementation) impls.getImplementation(i);
            unresolved.add(impl.getIdentifier());
        }

        while(unresolved.size() > 0)
        {
            impl = parent = (XMLImplementation) impls.getImplementation(
                (String) unresolved.get(unresolved.size() - 1));

            // Link the highest unresolved parent first.
            do
            {
                if(parent.getParent() != null)
                {
                    parent = (XMLImplementation) parent.getParent();
                    if(unresolved.contains(parent.getIdentifier()))
                    {
                        impl = parent;
                    }
                }
            } while(parent.getParent() != null);

            unresolved.remove(impl.getIdentifier());
            deps = impl.getDependencies();
            for(int j = deps.size() - 1; j >= 0; j--)
            {
                dep = (XMLDependency) deps.getDependency(j);
                spec = modules.getSpecifications().
                    getSpecification(dep.getSpecificationIdentifier());

                dep.setSpecification(spec);

                if(dep.getImplementationName() != null)
                {
                    try
                    {
                        dep.setImplementation(spec.getImplementation(
                            dep.getImplementationName()));

                    }
                    catch(MissingImplementationException e)
                    {
                        Logger.getLogger(DefaultModel.class.getName()).
                            log(Level.WARNING, e.getMessage() + ' ' +
                            DefaultModelBundle.getMissingImplementationMessage(
                            Locale.getDefault()).format(new Object[] {
                            impl.getIdentifier(), dep.getName() }));

                    }
                }
                else if(spec.getMultiplicity() ==
                    Specification.MULTIPLICITY_ONE)
                {
                    if(spec.getImplementations().size() > 1)
                    {
                        throw new MultiplicityConstraintException(
                            spec.getIdentifier());

                    }
                    else if(spec.getImplementations().size() == 0)
                    {
                        Logger.getLogger(DefaultModel.class.getName()).
                            log(Level.WARNING,
                            DefaultModelBundle.getUnresolvedImplementationText(
                            Locale.getDefault()) + ' ' +
                            DefaultModelBundle.getMissingImplementationMessage(
                            Locale.getDefault()).format(new Object[] {
                            impl.getIdentifier(), dep.getName() }));

                    }
                    else
                    {
                        dep.setImplementation(spec.getImplementations().
                            getImplementation(0));

                    }
                }
                else
                {
                    throw new MultiplicityConstraintException(
                        spec.getIdentifier());

                }
            }
        }

        // Checks the property overwrite constraint.
        impls = modules.getImplementations();
        for(int i = impls.size() - 1; i >= 0; i--)
        {
            deps = impls.getImplementation(i).getDependencies();
            for(int j = deps.size() - 1; j >= 0; j--)
            {
                dep = (XMLDependency) deps.getDependency(j);
                if(dep.getSpecification().isSingleton() &&
                    dep.isPropertiesDefined())
                {
                    throw new PropertyOverwriteConstraintException(
                        dep.getImplementation().getIdentifier(),
                        dep.getName());

                }
            }
        }

        // Checks the multiplicity constraint.
        specs = modules.getSpecifications();
        for(int i = specs.size() - 1; i >= 0; i--)
        {
            spec = specs.getSpecification(i);
            if(spec.getMultiplicity() == Specification.MULTIPLICITY_ONE &&
                spec.getImplementations().size() > 1)
            {
                throw new MultiplicityConstraintException(spec.getIdentifier());
            }
        }
    }

    private Modules transformModules(Collection xml)
    {
        int i;
        final Iterator it;
        final Modules mods = new Modules();
        final Module[] transformed = new Module[xml.size()];
        for(it = xml.iterator(), i = 0; it.hasNext(); i++)
        {
            transformed[i] = this.transformModule((Document) it.next());
        }

        mods.setModules(transformed);
        this.linkXMLModel(mods);
        return mods;
    }

    private Specifications transformSpecifications(final String modelVersion,
        final String moduleName, final NodeList xml)
    {

        final Specification[] transformed = new Specification[xml.getLength()];
        for(int i = xml.getLength() - 1; i >= 0; i--)
        {
            transformed[i] = this.transformSpecification(modelVersion,
                moduleName, (Element) xml.item(i));

        }

        final Specifications specs = new Specifications();
        specs.setSpecifications(transformed);
        return specs;
    }

    private Implementations transformImplementations(final String modelVersion,
        final String moduleName, final NodeList xml)
    {

        final Implementation[] transformed =
            new Implementation[xml.getLength()];

        for(int i = xml.getLength() - 1; i >= 0; i--)
        {
            transformed[i] = this.transformImplementation(modelVersion,
                moduleName, (Element) xml.item(i));

        }

        final Implementations impls = new Implementations();
        impls.setImplementations(transformed);
        return impls;
    }

    private Properties transformProperties(final NodeList xml)
    {
        final Property[] transformed = new Property[xml.getLength()];
        for(int i = xml.getLength() - 1; i >= 0; i--)
        {
            transformed[i] = this.transformProperty((Element) xml.item(i));
        }

        final Properties props = new Properties();
        props.setProperties(transformed);
        return props;
    }

    private Dependencies transformDependencies(final NodeList xml)
    {
        final Dependency[] transformed = new Dependency[xml.getLength()];
        for(int i = xml.getLength() - 1; i >= 0; i--)
        {
            transformed[i] = this.transformDependency((Element) xml.item(i));
        }

        final Dependencies deps = new Dependencies();
        deps.setDependencies(transformed);
        return deps;
    }

    private String[] transformImplementedSpecifications(final NodeList xml)
    {
        final String[] transformed = new String[xml.getLength()];
        for(int i = xml.getLength() - 1; i >= 0; i--)
        {
            transformed[i] = this.transformImplementedSpecification(
                (Element) xml.item(i));

        }

        return transformed;
    }

    private Module transformModule(Document xml)
    {
        NodeList l;
        final Module module = new Module();
        final Element e = xml.getDocumentElement();
        module.setName(e.getAttributeNS(DefaultModel.MODEL_NS, "name"));
        module.setDescription(e.getAttributeNS(DefaultModel.MODEL_NS,
            "description"));

        module.setVersion(e.getAttributeNS(DefaultModel.MODEL_NS, "version"));
        l = e.getElementsByTagNameNS(DefaultModel.MODEL_NS,
            "specifications");

        if(l != null && l.getLength() > 0)
        {
            module.setSpecifications(this.transformSpecifications(
                module.getVersion(), module.getName(),
                ((Element) l.item(0)).getElementsByTagNameNS(
                DefaultModel.MODEL_NS, "specification")));

        }

        l = e.getElementsByTagNameNS(DefaultModel.MODEL_NS,
            "implementations");

        if(l != null && l.getLength() > 0)
        {
            module.setImplementations(this.transformImplementations(
                module.getVersion(), module.getName(),
                ((Element) l.item(0)).getElementsByTagNameNS(
                DefaultModel.MODEL_NS, "implementation")));

        }

        l = e.getElementsByTagNameNS(DefaultModel.MODEL_NS,
            "properties");

        if(l != null && l.getLength() > 0)
        {
            // Find the module properties.
            for(int i = l.getLength() - 1; i >= 0; i--)
            {
                if("module".equals(l.item(i).getParentNode().getLocalName()))
                {
                    module.setProperties(this.transformProperties(
                        ((Element) l.item(i)).getElementsByTagNameNS(
                        DefaultModel.MODEL_NS, "properties")));

                    break;
                }
            }

        }

        if(module.getVersion() != null)
        {
            boolean versionSupported = false;

            for(int i = SUPPORTED_VERSIONS.length - 1; i >= 0; i--)
            {
                if(SUPPORTED_VERSIONS[i].equals(module.getVersion()))
                {
                    versionSupported = true;
                    break;
                }
            }

            if(!versionSupported)
            {
                throw new ModelError(module.getVersion());
            }
        }
        else
        {
            throw new ModelError(module.getVersion());
        }

        return module;
    }

    private Specification transformSpecification(final String modelVersion,
        final String moduleName, final Element xml)
    {
        String attr;
        final Specification spec = new Specification();
        spec.setDescription(xml.getAttributeNS(DefaultModel.MODEL_NS,
            "description"));

        spec.setIdentifier(xml.getAttributeNS(DefaultModel.MODEL_NS,
            "identifier"));

        spec.setVendor(xml.getAttributeNS(DefaultModel.MODEL_NS, "vendor"));
        spec.setVersion(xml.getAttributeNS(DefaultModel.MODEL_NS, "version"));

        spec.setModuleName(moduleName);
        attr = xml.getAttributeNS(DefaultModel.MODEL_NS, "singleton");
        spec.setSingleton(Boolean.valueOf(attr).booleanValue());

        if(modelVersion.equals("1.1"))
        {
            attr = xml.getAttributeNS(DefaultModel.MODEL_NS, "multiplicity");
            if("one".equalsIgnoreCase(attr))
            {
                spec.setMultiplicity(Specification.MULTIPLICITY_ONE);
            }
            else if("many".equalsIgnoreCase(attr))
            {
                spec.setMultiplicity(Specification.MULTIPLICITY_MANY);
            }
            else
            {
                throw new ModelError(attr);
            }
        }

        return spec;
    }

    private Implementation transformImplementation(final String modelVersion,
        final String moduleName, final Element xml)
    {
        NodeList l;
        final XMLImplementation impl = new XMLImplementation();
        impl.setIdentifier(xml.getAttributeNS(DefaultModel.MODEL_NS,
            "identifier"));

        impl.setModuleName(moduleName);
        impl.setName(xml.getAttributeNS(DefaultModel.MODEL_NS, "name"));
        impl.setParentIdentifier(xml.getAttributeNS(DefaultModel.MODEL_NS,
            "parent"));

        impl.setVendor(xml.getAttributeNS(DefaultModel.MODEL_NS, "vendor"));
        impl.setVersion(xml.getAttributeNS(DefaultModel.MODEL_NS, "version"));

        l = xml.getElementsByTagNameNS(DefaultModel.MODEL_NS,
            "dependencies");

        if(l != null && l.getLength() > 0)
        {
            impl.setDependencies(this.transformDependencies(
                ((Element) l.item(0)).getElementsByTagNameNS(
                DefaultModel.MODEL_NS, "dependency")));

        }

        l = xml.getElementsByTagNameNS(DefaultModel.MODEL_NS,
            "implementedSpecifications");

        if(l != null && l.getLength() > 0)
        {
            impl.setImplementedIdentifiers(
                this.transformImplementedSpecifications(((Element) l.item(0)).
                getElementsByTagNameNS(DefaultModel.MODEL_NS,
                "implementedSpecification")));

        }

        l = xml.getElementsByTagNameNS(DefaultModel.MODEL_NS,
            "properties");

        if(l != null && l.getLength() > 0)
        {
            // Find the implementation properties.
            for(int i = l.getLength() - 1; i >= 0; i--)
            {
                if("implementation".equals(l.item(i).getParentNode().
                    getLocalName()))
                {

                    impl.setProperties(this.transformProperties(
                        ((Element) l.item(i)).getElementsByTagNameNS(
                        DefaultModel.MODEL_NS, "property")));

                    break;
                }
            }
        }

        if(modelVersion.equals("1.1"))
        {
            final String attr =
                xml.getAttributeNS(DefaultModel.MODEL_NS, "final");

            if(attr != null && attr.length() > 0)
            {
                impl.setFinal(Boolean.valueOf(attr).booleanValue());
            }
        }

        return impl;
    }

    private Property transformProperty(final Element xml)
    {
        String attr;
        String t;
        Class objectType = null;
        Class type = null;
        final Property prop = new Property();
        prop.setName(xml.getAttributeNS(DefaultModel.MODEL_NS, "name"));

        attr = xml.getAttributeNS(DefaultModel.MODEL_NS, "api");
        prop.setApi(attr != null ?
            Boolean.valueOf(attr).booleanValue() : false);

        t = xml.getAttributeNS(DefaultModel.MODEL_NS, "type");
        if(t.equals(Boolean.TYPE.getName()))
        {
            type = Boolean.TYPE;
            objectType = Boolean.class;
        }
        else if(t.equals(Byte.TYPE.getName()))
        {
            type = Byte.TYPE;
            objectType = Byte.class;
        }
        else if(t.equals(Character.TYPE.getName()))
        {
            type = Character.TYPE;
            objectType = Character.class;
        }
        else if(t.equals(Double.TYPE.getName()))
        {
            type = Double.TYPE;
            objectType = Double.class;
        }
        else if(t.equals(Float.TYPE.getName()))
        {
            type = Float.TYPE;
            objectType = Float.class;
        }
        else if(t.equals(Integer.TYPE.getName()))
        {
            type = Integer.TYPE;
            objectType = Integer.class;
        }
        else if(t.equals(Long.TYPE.getName()))
        {
            type = Long.TYPE;
            objectType = Long.class;
        }
        else if(t.equals(Short.TYPE.getName()))
        {
            type = Short.TYPE;
            objectType = Short.class;
        }
        else if(t.equals(Boolean.class.getName()))
        {
            type = objectType = Boolean.class;
        }
        else if(t.equals(Byte.class.getName()))
        {
            type = objectType = Byte.class;
        }
        else if(t.equals(Character.class.getName()))
        {
            type = objectType = Character.class;
        }
        else if(t.equals(Double.class.getName()))
        {
            type = objectType = Double.class;
        }
        else if(t.equals(Float.class.getName()))
        {
            type = objectType = Float.class;
        }
        else if(t.equals(Integer.class.getName()))
        {
            type = objectType = Integer.class;
        }
        else if(t.equals(Long.class.getName()))
        {
            type = objectType = Long.class;
        }
        else if(t.equals(Short.class.getName()))
        {
            type = objectType = Short.class;
        }
        else if(t.equals(String.class.getName()))
        {
            type = objectType = String.class;
        }
        else
        {
            throw new ModelError(t);
        }
        prop.setType(type);

        attr = xml.getAttributeNS(DefaultModel.MODEL_NS, "value");
        try
        {
            // Special handling for class Character which does not provide
            // a constructor taking a string.
            final Constructor ctor;
            final Object arg;

            if(objectType == Character.class)
            {
                ctor = objectType.getDeclaredConstructor(
                    DefaultModel.CTOR_ARGS_CHAR);

                if(attr.length() != 1)
                {
                    throw new PropertyException(prop.getName(), attr);
                }

                arg = new Character(attr.charAt(0));
            }
            else
            {
                ctor = objectType.getDeclaredConstructor(
                    DefaultModel.CTOR_ARGS_STRING);

                arg = attr;
            }

            prop.setValue(ctor.newInstance(new Object[] { arg }));
        }
        catch (SecurityException e)
        {
            throw new ModelError(e);
        }
        catch (NoSuchMethodException e)
        {
            throw new ModelError(e);
        }
        catch (InvocationTargetException e)
        {
            throw new ModelError(e.getTargetException() == null ?
                e : e.getTargetException());

        }
        catch (IllegalAccessException e)
        {
            throw new ModelError(e);
        }
        catch (InstantiationException e)
        {
            throw new ModelError(e);
        }

        return prop;
    }

    private Dependency transformDependency(final Element xml)
    {
        String attr;
        NodeList l;
        final XMLDependency dep = new XMLDependency();
        dep.setName(xml.getAttributeNS(DefaultModel.MODEL_NS, "name"));
        attr = xml.getAttributeNS(DefaultModel.MODEL_NS, "bound");
        dep.setBound(attr != null ?
            Boolean.valueOf(attr).booleanValue() : false);

        attr = xml.getAttributeNS(DefaultModel.MODEL_NS, "implementationName");
        dep.setImplementationName(
            attr != null && attr.length() > 0 ? attr : null);

        dep.setSpecificationIdentifier(xml.getAttributeNS(DefaultModel.MODEL_NS,
            "specificationIdentifier"));

        l = xml.getElementsByTagNameNS(DefaultModel.MODEL_NS,
            "properties");

        if(l != null && l.getLength() > 0)
        {
            dep.setProperties(this.transformProperties(
                ((Element) l.item(0)).getElementsByTagNameNS(
                DefaultModel.MODEL_NS, "property")));

            dep.setPropertiesDefined(true);
        }

        return dep;
    }

    private String transformImplementedSpecification(final Element xml)
    {
        return xml.getAttributeNS(DefaultModel.MODEL_NS, "identifier");
    }

    //------------------------------------------------------------DefaultModel--
}
